# -*- coding: future_fstrings -*-
from typing import *
from functools import reduce
from collections import defaultdict, OrderedDict
from collections import UserList
import pandas as pd
from copy import copy

ReducerT = Callable[[pd.Series], Any]
StrOrTupleT = Union[str, tuple]


class QuerySet(UserList):
    def __repr__(self):
        return f"QuerySet({repr(self.data)})"

    def project(self, on: List[Union[StrOrTupleT, Dict[StrOrTupleT, ReducerT]]], rename="last") -> pd.DataFrame:
        """Project a set of experiments onto a dataframe.

        Parameters
        ----------
        on: List of dot separated paths that point to values in the experiment model. Instead of a dot separated path
            passing a tuple is also possible. This allows for pointing to path in the model that contain dots themself.
             Additionally, a dict of a path mapping to function applied to the
            values in the experiment model can be used inplace of a path. This is useful for summarizing metrics.
        rename: The renaming strategy used to create the column names. Either "last" to take the last element in each
                path as a column name or None to use the complete paths as column names.

        Returns
        -------
        A dataframe containing the projected values with the experiment id set as index.
        """
        on = self._stratify_mapping(on)

        # TODO introduce possibility to pass a list to `rename` once we don't need to support 3.5 any longer.
        rename_mapping = self._make_rename_mapping(on, rename)

        projected = defaultdict(list)
        for exp in self.data:
            projected["exp_id"].append(exp.id)
            for path, reducer in on.items():
                projected[path].append(self._extract(exp, path, reducer))

        return pd.DataFrame(projected).set_index("exp_id").rename(columns=rename_mapping)

    def _make_rename_mapping(self, on, rename):
        rename_mapping = {}
        for path, reducer_or_path in on.items():
            if rename == "last":
                str_path = path[-1]
            elif rename is None:
                str_path = '.'.join(path)

            if callable(reducer_or_path):
                rename_mapping[path] = f"{str_path}_{reducer_or_path.__name__}"
            else:
                rename_mapping[path] = str_path
        return rename_mapping

    def _stratify_mapping(self, on):
        stratified_on = OrderedDict()
        for path in on:
            if isinstance(path, dict):
                stratified_on.update(path)
            else:
                stratified_on[path] = path
        on = stratified_on

        for path, value in copy(on).items():
            del on[path]
            if isinstance(path, str):
                path = tuple(path.split("."))
            on[path] = value

        return on

    def _extract(self, exp, path, name):
        extracted = reduce(lambda x, y: self._get(x, y), path, exp)
        if isinstance(name, str):
            return extracted
        elif callable(name):
            return name(extracted)

    def _get(self, o, name):
        """Try getattr and getitem."""
        try:
            return getattr(o, name)
        except AttributeError:
            return o[name]



# -*- coding: future_fstrings -*-
import pickle
import os
from IPython.display import HTML
from IPython import display
import pandas as pd
import warnings
from copy import copy


class Artifact:
    """Displays or saves an artifact."""

    can_render = tuple()

    def __init__(self, name: str, file, content_type: str = None):
        self.name = name
        self.file = file
        self.content_type = content_type
        self.extension = None if self.content_type is None else self.content_type.split("/")[-1]
        self._content = None
        self._rendered = None

    def __repr__(self):
        return f'{self.__class__.__name__}(name={self.name})'

    def render(self):
        """Render the artifact according to its content-type."""
        if self._rendered is None:
            self._rendered = self._render()
        return self._rendered

    def _render(self):
        """Return the object that represents the rendered artifact."""
        raise NotImplementedError

    def show(self):
        warnings.warn("`show` is deprecated in favor of `render` and will removed in a future release.",
                      DeprecationWarning,
                      stacklevel=2)
        return self.render()

    def save(self, save_dir: str = ''):
        """Save artifact to disk."""
        with open(os.path.join(save_dir, self._make_filename()), 'wb') as file:
            file.write(self.content)

    def as_content_type(self, content_type):
        """Interpret artifact as being of content-type."""
        try:
            artifact_type = content_type_to_artifact_cls[content_type]
        except KeyError:
            raise ValueError(f'Incense does not have a class that maps to content-type {content_type}')
        else:
            return self.as_type(artifact_type)

    def as_type(self, artifact_type):
        self.file.seek(0)
        return artifact_type(self.name, self.file)

    @property
    def content(self):
        """Access the raw bytes of the artifact."""
        if self._content is None:
            self._content = self.file.read()
        return self._content

    def _make_filename(self):
        parts = self.file.filename.split('/')
        return f'{parts[-2]}_{parts[-1]}.{self.extension}'


class ImageArtifact(Artifact):
    """Displays or saves an image artifact."""

    can_render = ('image/png', 'image/jpeg')

    def _render(self):
        return display.Image(data=self.content)


class MP4Artifact(Artifact):
    """Displays or saves a MP4 artifact"""

    can_render = ("video/mp4",)

    def _render(self):
        self.save()
        return HTML(f"""
        <video width="640" height="480" controls autoplay>
          <source src="{self._make_filename()}" type="video/mp4">
        </video>
        """)


class CSVArtifact(Artifact):
    """Displays and saves a CSV artifact"""

    can_render = ("text/csv",)

    def _render(self):
        return pd.read_csv(self.file)


class PickleArtifact(Artifact):
    """Displays and saves a Pickle artifact"""

    can_render = tuple()

    def __init__(self, name: str, file, content_type: str = None):
        super().__init__(name, file, content_type)
        self.extension = "pickle"

    def _render(self):
        return pickle.load(self.file)


class PDFArtifact(Artifact):
    """Displays and saves a PDF artifacts."""
    can_render = ("application/pdf",)

    # TODO probably needs jupyter extension to be able to display pdf.
    # def _render(self):
    #     return IFrame(self._make_filename(), width=600, height=300)
    #


content_type_to_artifact_cls = {}
for cls in copy(locals()).values():
    # print(cls)
    if isinstance(cls, type) and issubclass(cls, Artifact):
        for content_type in cls.can_render:
            content_type_to_artifact_cls[content_type] = cls

# print(content_type_to_artifact_cls)
